/*
 * Copyright 2010 by egnite GmbH
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

#include "config.h"

#include <arch/arm.h>
#include <stdio.h>

#include "at45d.h"

#define DFCMD_READ_STATUS       0xD7
#define DFCMD_PAGE_ERASE        0x81
#define DFCMD_CONT_READ         0xE8

#define CS_ENABLE(x)    outr(DF_CS_PIO + PIO_CODR_OFF, _BV(x));
#define CS_DISABLE(x)   outr(DF_CS_PIO + PIO_SODR_OFF, _BV(x));

typedef struct _AT45_DEVTAB {
    uint32_t devt_pages;
    unsigned int devt_pagsiz;
    unsigned int devt_offs;
    uint8_t devt_srmsk;
    uint8_t devt_srval;
} AT45_DEVTAB;

#define DEVTAB_ENTRIES  8

static AT45_DEVTAB at45_devt[DEVTAB_ENTRIES] = {
    {2048, 264, 9, 0x3C, 0x1C},     /* AT45DB041  - 512kB */
    {4096, 264, 9, 0x3C, 0x24},     /* AT45DB081  -  1MB */
    {4096, 528, 10, 0x3C, 0x2C},    /* AT45DB0161 -  2MB */
    {8192, 528, 10, 0x3C, 0x34},    /* AT45DB0321 -  4MB */
    {8192, 1056, 11, 0x38, 0x38},   /* AT45DB0642 -  8MB */
    {16384, 1056, 11, 0x38, 0x38},  /* AT45DB1282 - 16MB */
    {16384, 2112, 12, 0x38, 0x18},  /* AT45DB2562 - 32MB */
    {32768, 2112, 12, 0x38, 0x20}   /* AT45DB5122 - 64MB */
};

static uint_fast8_t df_type;

/*!
 * \brief Initialize hardware SPI.
 */
static void SpiInit(void)
{
    /* Enable chip selects, using GPIO. */
    outr(DF_CS_PIO + PIO_PER_OFF, _BV(DF_CS_BIT));
    outr(DF_CS_PIO + PIO_SODR_OFF, _BV(DF_CS_BIT));
    outr(DF_CS_PIO + PIO_OER_OFF, _BV(DF_CS_BIT));

#if defined(AT91SAM9260_EK)
    /* Enable peripherals. */
    outr(PIOA_PDR, _BV(PA0_SPI0_MISO_A) | _BV(PA1_SPI0_MOSI_A) | _BV(PA2_SPI0_SPCK_A));
    outr(PIOA_ASR, _BV(PA0_SPI0_MISO_A) | _BV(PA1_SPI0_MOSI_A) | _BV(PA2_SPI0_SPCK_A));
    /* Enable SPI. */
    outr(SPI0_CR, SPI_SPIEN);
    /* Set SPI mode. */
    outr(SPI0_MR, (1 << SPI_PCS_LSB) | SPI_MODFDIS | SPI_MSTR);
    outr(SPI0_CSR1, /* SPI_CPOL | */ SPI_NCPHA | (4 << SPI_SCBR_LSB));
#else
#error SPI initialization missing
#endif
}

/*!
 * \brief Send a byte and receive a byte via SPI.
 *
 * Note, that this function will enable the chip select.
 *
 * \param c The byte to send.
 */
static uint8_t SpiByte(uint8_t c)
{
    CS_ENABLE(DF_CS_BIT);
#if defined(AT91SAM9260_EK)
    /* Transmission starts by writing the transmit data. */
    outr(SPI0_TDR, c);
    /* Wait for receiver data register full. */
    while((inr(SPI0_SR) & SPI_RDRF) == 0);
    /* Read incoming data. */
    return (uint8_t)inr(SPI0_RDR);
#else
#error Undefined SPI transfer
#endif
}

/*!
 * \brief Send 32-bit value to device.
 *
 * The most significant byte is sent first. Note, that this function 
 * calls SpiByte(), which will enable the chip select.
 *
 * \param v The value to send.
 */
static void SpiSendLong(uint32_t v)
{
    SpiByte((uint8_t)(v >> 24));
    SpiByte((uint8_t)(v >> 16));
    SpiByte((uint8_t)(v >> 8));
    SpiByte((uint8_t)v);
}

/*!
 * \brief Get the DataFlash status byte.
 *
 * \return The status byte.
 */
uint8_t DataFlashStatus(void)
{
    uint8_t rc;

    SpiByte(DFCMD_READ_STATUS);
    rc = SpiByte(0xFF);
    CS_DISABLE(DF_CS_BIT);

    return rc;
}

/*!
 * \brief Initialize DataFlash.
 *
 * This function will enable the SPI and determine the DataFlash type.
 * The following devices are currently supported:
 * - AT45DB041
 * - AT45DB081
 * - AT45DB0161
 * - AT45DB0321
 * - AT45DB0642
 * - AT45DB1282
 * - AT45DB2562
 * - AT45DB5122
 *
 * \return 0 on success, -1 if the function was unable to determine the
 *         DataFlash type.
 */
int DataFlashInit(void)
{
    uint8_t sr;

    SpiInit();
    sr = DataFlashStatus();
    for (df_type = 0; df_type < DEVTAB_ENTRIES; df_type++) {
        if ((sr & at45_devt[df_type].devt_srmsk) == at45_devt[df_type].devt_srval) {
            return 0;
        }
    }
    return -1;
}

/*!
 * \brief Recover from bad image.
 *
 * When calling this function, the first DataFlash page will be erased.
 * This will re-enable the SAM-BA ROM code on the next system start.
 */
void DataFlashRecover(void)
{
    SpiSendLong(DFCMD_PAGE_ERASE << 24);
    CS_DISABLE(DF_CS_BIT);

    while ((DataFlashStatus() &0x80) == 0);
}

/*!
 * \brief Read image from the DataFlash device.
 *
 * \param first DataFlash location to start reading.
 * \param mem   Memory location to store the image.
 * \param len   Size of the image.
 */
void DataFlashRead(uint32_t first, uint8_t *mem, size_t len)
{
    size_t i;

    first /= at45_devt[df_type].devt_pagsiz;
    first <<= at45_devt[df_type].devt_offs;

    SpiSendLong((DFCMD_CONT_READ << 24) | first);
    SpiSendLong(0xFFFFFFFF);
    for (i = 0; i < len; i++) {
        *mem++ = SpiByte(0xFF);
    }
    CS_DISABLE(DF_CS_BIT);
}
