/*
 * Copyright 2010 by egnite GmbH
 * Copyright 2001-2005 by egnite Software GmbH
 * Copyright 1990, 1993 by The Regents of the University of California
 *
 * All rights reserved.
 *
 * This code is partly derived from software contributed to Berkeley by
 * Chris Torek, but heavily rewritten for SAMBOOT.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

#include "config.h"

#if UART_OUTPUT

#include <arch/arm.h>
#include <stdint.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>

#include "utils.h"

#define MAX_NUMLEN     12

#define FMTFLG_ALT		0x01
#define FMTFLG_LADJ     0x04
#define FMTFLG_LONG     0x08
#define FMTFLG_ZPAD     0x10

static const char xdigits[16] = {
'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
};

static int initialized;

static const char *atodec(const char *cp, int *ip)
{
    *ip = 0;
    while (*cp >= '0' && *cp <= '9') {
        *ip *= 10;
        *ip += *cp++;
        *ip -= '0';
    }
    return cp;
}

/*!
 * \brief Print a character.
 *
 * \return Always returns the character, which differs from the 
 *         standard implementation.
 */
int putchar(int c)
{
    if (!initialized) {
        outr(DBGU_BRGR, BAUDRATE_DIV);
        outr(PIOB_PDR, _BV(PB15_DTXD_A));
        outr(DBGU_MR, US_CHMODE_NORMAL | US_CHRL_8 | US_PAR_NO | US_NBSTOP_1);
        outr(DBGU_CR, US_TXEN);
        initialized = 1;
    }
    if (c == '\n') {
        putchar('\r');
    }
    while ((inr(DBGU_SR) & US_TXRDY) == 0);
    outr(DBGU_THR, c);

    return c;
}

static void print(const char *s, int len)
{
    while (len-- > 0) {
        putchar(*s);
        s++;
    }
}

static void printm(int c, int len)
{
    while (len-- > 0) {
        putchar(c);
    }
}

/*!
 * \brief Print a string.
 *
 * \return Always returns 0, which differs from the standard implementation.
 */
int puts(const char *s)
{
    while (*s) {
        putchar(*s);
        s++;
    }
    putchar('\n');
    return 0;
}

/*!
 * \brief Formatted output with argument list.
 *
 * The following types are supported:
 * - 'd' for signed decimal numbers.
 * - 'u' for unsigned decimal numbers.
 * - 'x' and 'X' for hexadecimal numbers, uppercase only.
 * - 'c' for characters.
 * - 's' for strings.
 *
 * In addition, the standard modifiers listed below are available.
 *
 * Flags:
 * - Numbers for padding length, including leading zero for '0' padding.
 * - '+' for forcing sign.
 * - '-' for left alignment.
 * - '#' for leading '0x' of hexadecimal numbers.
 *
 * Precision:
 * - '.' followed by a number to limit string lengths.
 * - '*' to read precision from the next parameter.
 *
 * Length specifier:
 * - 'l' for long values.
 * - 'z' for size_t values.
 *
 * \param fmt Format string.
 * \param ap  List of arguments.
 *
 * \return Always returns 0, which differs from the standard implementation.
 */
int vprintf(const char *fmt, va_list ap)
{
    int_fast8_t ch;		/* Character from fmt. */
    int n;              /* Handy integer (short term usage). */
    char *cp;		    /* Handy char pointer (short term usage). */
    uint_fast8_t flags; /* Flags as above. */
    int_fast8_t width;  /* Width from format (%8d), or 0. */
    int_fast8_t prec;   /* Precision from format (%.3d), or -1. */
    int_fast8_t dprec;  /* A copy of prec if [duxX], 0 otherwise. */
    int_fast8_t realsz; /* Field size expanded by dprec, sign, etc. */
    char sign;          /* Sign prefix ('+', '-', or '\0'). */
    unsigned long ulval; /* Integer arguments %[duxX]. */
    int_fast8_t size;   /* Size of converted field or string. */
    char buf[MAX_NUMLEN]; /* Space for %[cduxX]. */

    for (;;) {

        /*
         * Print format string until next percent sign.
         */
        for (n = 0; fmt[n] && fmt[n] != '%'; n++);
        print(fmt, n);
        if (fmt[n] == '\0') {
            break;
        }
        fmt += n + 1;

        /*
         * Process modifiers.
         */
        flags = 0;
        sign = '\0';
        width = 0;
        dprec = 0;
        prec = -1;
        for (;;) {
            ch = *fmt++;
            if (ch == '+') {
                sign = '+';
            }
            else if (ch == '-') {
                flags |= FMTFLG_LADJ;
            }
            else if (ch == '#') {
                flags |= FMTFLG_ALT;
            }
            else if (ch == '0') {
                flags |= FMTFLG_ZPAD;
            }
            else if (ch == 'l') {
                flags |= FMTFLG_LONG;
            }
            else if (ch == 'z') {
                if (sizeof(size_t) > sizeof(int)) {
                    flags |= FMTFLG_LONG;
                }
            }
            else if (ch == '*') {
                width = va_arg(ap, int);
                if (width < 0) {
                    flags |= FMTFLG_LADJ;
                    width = -width;
                }
            } 
            else if (ch == '.') {
                if (*fmt == '*') {
                    fmt++;
                    prec = va_arg(ap, int);
                } else {
                    fmt = atodec(fmt, &prec);
                }
                if (prec < 0) {
                    prec = -1;
                }
            } 
            else if (ch >= '1' && ch <= '9') {
                fmt = atodec(fmt - 1, &width);
            } 
            else {
                break;
            }
        }

        /*
         * Process type field.
         */
        cp = buf;
        size = 1;
        if (ch == 'c') {
            *cp = va_arg(ap, int);
        }
        else if (ch == 's') {
            cp = va_arg(ap, char *);
            if (cp == NULL) {
                cp = "(null)";
            }
            size = strlen(cp);
            if (prec >= 0 && size > prec) {
                size = prec;
            }
        }
        else if (ch == 'd' || ch == 'u' || ch == 'x' || ch == 'X') {
            ulval = (flags & FMTFLG_LONG) ? va_arg(ap, unsigned long) : va_arg(ap, unsigned int);
            if (ch == 'd') {
                if ((long) ulval < 0) {
                    ulval = (unsigned long) (-((long) ulval));
                    sign = '-';
                }
            }
            dprec = prec;
            if (dprec >= 0) {
                flags &= ~FMTFLG_ZPAD;
            }

            cp = buf + MAX_NUMLEN;
            if (ch =='d' || ch == 'u') {
                if (ulval || prec) {
                    if (ulval < 10) {
                        *--cp = (char) ulval + '0';
                    }
                    else do {
                        *--cp = (char) (ulval % 10) + '0';
                        ulval /= 10;
                    } while (ulval);
                }
            } else {
                do {
                    *--cp = xdigits[ulval & 0x0f];
                    ulval >>= 4;
                } while (ulval);
                if (flags & FMTFLG_ALT) {
                    *--cp = (char)ch;
                    *--cp = '0';
                }
            }
            size = (buf + MAX_NUMLEN) - cp;
        }
        else {
            ulval = va_arg(ap, unsigned int);
            if (ch == 0) {
                return 0;
            }
            *cp = (char)ch;
        }

        /*
         * Output.
         */
        realsz = dprec > size ? dprec : size;
        if (sign) {
            realsz++;
        }
        if ((flags & (FMTFLG_LADJ | FMTFLG_ZPAD)) == 0) {
            printm(' ', width - realsz);
        }
        if (sign) {
            putchar(sign);
        }
        if ((flags & (FMTFLG_LADJ | FMTFLG_ZPAD)) == FMTFLG_ZPAD) {
            printm('0', width - realsz);
        }
        printm('0', dprec - size);

        if (size) {
            print(cp, size);
        }
        if (flags & FMTFLG_LADJ) {
            printm(' ', width - realsz);
        }
    }
    return 0;
}

/*!
 * \brief Formatted output.
 *
 * Note, that not all standard formatting features may be available. See
 * vprintf() for further details.
 *
 * \return Always returns 0, which differs from the standard implementation.
 */
int printf(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    vprintf(fmt, ap);
    va_end(ap);
    return 0;
}

/* UART_OUTPUT */
#endif
